from bs4 import BeautifulSoup
from pathlib import Path
from tqdm import tqdm
from typing import Any

from mixxx_bisect.utils.version import pkg_version

import functools
import requests
import shutil

def get(url: str, **kwargs: Any) -> requests.Response:
    headers = {'User-Agent': f'mixxx-bisect/{pkg_version()}'}
    response = requests.get(url, headers=headers, **kwargs)
    response.raise_for_status()
    return response

def download(url: str, output: Path):
    # https://stackoverflow.com/a/63831344
    response = get(url, stream=True, allow_redirects=True)
    file_size = int(response.headers.get('Content-Length', 0))

    # Decompress if needed
    response.raw.read = functools.partial(response.raw.read, decode_content=True)

    with tqdm.wrapattr(response.raw, 'read', total=file_size) as raw: # type: ignore
        with output.open('wb') as f:
            shutil.copyfileobj(raw, f)

def get_soup(url: str) -> BeautifulSoup:
    raw = get(url).content
    return BeautifulSoup(raw, 'html.parser')



from typing import Protocol

from mixxx_bisect.options import Options

class SnapshotHoster(Protocol):
    '''A snapshot archive hosted on the web.'''

    def __init__(self, branch: str, suffix: str, opts: Options):
        pass

    def fetch_snapshots(self) -> dict[str, str]:
        '''Fetches the snapshots as a dictionary from commit SHAs to URLs..'''
        raise NotImplementedError()



from pathlib import Path

from mixxx_bisect.runner import SnapshotRunner
from mixxx_bisect.options import Options
from mixxx_bisect.utils.run import run

class MacOSSnapshotRunner(SnapshotRunner):
    def __init__(self, opts: Options):
        self.mount_dir = opts.mount_dir
        self.opts = opts
    
    @property
    def download_path(self) -> Path:
        return self.opts.downloads_dir / 'mixxx-current.dmg'
    
    @property
    def cdr_path(self) -> Path:
        return self.download_path.with_suffix('.cdr')

    def _mount_snapshot(self):
        print('Mounting snapshot...')
        self.cdr_path.unlink(missing_ok=True)
        run(['hdiutil', 'convert', str(self.download_path), '-format', 'UDTO', '-o', str(self.cdr_path)], opts=self.opts)
        run(['hdiutil', 'attach', str(self.cdr_path), '-mountpoint', str(self.mount_dir)], opts=self.opts)

    def _unmount_snapshot(self):
        print('Unmounting snapshot...')
        run(['hdiutil', 'unmount', str(self.mount_dir)], cwd=self.mount_dir.parent, opts=self.opts)

    def _delete_snapshot(self):
        print('Deleting snapshot...')
        self.download_path.unlink(missing_ok=True)
        self.cdr_path.unlink(missing_ok=True)

    def setup_snapshot(self):
        self._mount_snapshot()

    def run_snapshot(self):
        print('Running snapshot...')
        run([str(self.mount_dir / 'mixxx.app' / 'Contents' / 'MacOS' / 'mixxx')], opts=self.opts)
    
    def cleanup_snapshot(self):
        self._unmount_snapshot()
        self._delete_snapshot()



from pathlib import Path
from typing import Protocol

from mixxx_bisect.options import Options

class SnapshotRunner(Protocol):
    '''A platform-dependent facility for extracting and running snapshots.'''

    def __init__(self, opts: Options):
        pass

    @property
    def download_path(self) -> Path:
        '''The path to download to.'''
        raise NotImplementedError()

    def setup_snapshot(self) -> None:
        '''Extracts or mounts the snapshot.'''
        raise NotImplementedError()
    
    def run_snapshot(self) -> None:
        '''Runs the snapshot.'''
        raise NotImplementedError()
    
    def cleanup_snapshot(self) -> None:
        '''Cleans up the snapshot.'''
        raise NotImplementedError()



from dataclasses import dataclass
from pathlib import Path

@dataclass
class Options:
    quiet: bool
    root_dir: Path
    mixxx_dir: Path
    mount_dir: Path
    log_dir: Path
    downloads_dir: Path



import argparse
import json
import platform
import sys

from pathlib import Path
from mixxx_bisect.hoster import SnapshotHoster
from mixxx_bisect.hoster.m1xxx import M1xxxSnapshotHoster
from mixxx_bisect.hoster.mixxx_org import MixxxOrgSnapshotHoster

from mixxx_bisect.options import Options
from mixxx_bisect.runner import SnapshotRunner
from mixxx_bisect.runner.macos import MacOSSnapshotRunner
from mixxx_bisect.runner.windows import WindowsSnapshotRunner
from mixxx_bisect.utils.git import clone_mixxx, commits_in_order, describe_commit, parse_commit, sort_commits
from mixxx_bisect.utils.snapshot import download_snapshot
from mixxx_bisect.utils.version import pkg_version

DEFAULT_ROOT = Path.home() / '.local' / 'state' / 'mixxx-bisect'
OS = platform.system()

# Platform-specific snapshot runners

SNAPSHOT_RUNNERS: dict[str, type[SnapshotRunner]] = {
    'Windows': WindowsSnapshotRunner,
    'Darwin': MacOSSnapshotRunner,
}

SNAPSHOT_HOSTERS: dict[str, type[SnapshotHoster]] = {
    'mixxx-org': MixxxOrgSnapshotHoster,
    'm1xxx': M1xxxSnapshotHoster,
}

# Main

def main():
    if OS not in SNAPSHOT_RUNNERS.keys():
        print(f"Unsupported OS: {OS} has no snapshot runner (supported are {', '.join(SNAPSHOT_RUNNERS.keys())})")
        sys.exit(1)

    SnapshotRunner = SNAPSHOT_RUNNERS[OS]

    parser = argparse.ArgumentParser(description='Finds Mixxx regressions using binary search')
    parser.add_argument('--hoster', default='mixxx-org', choices=sorted(SNAPSHOT_HOSTERS.keys()), help=f'THe snapshot archive to use.')
    parser.add_argument('--branch', default='main', help=f'The branch to search for snapshots on, if supported by the hoster.')
    parser.add_argument('--root', type=Path, default=DEFAULT_ROOT, help='The root directory where all application-specific state (i.e. the mixxx repo, downloads, mounted snapshots etc.) will be stored.')
    parser.add_argument('--dump-snapshots', action='store_true', help='Dumps the fetched snapshots.')
    parser.add_argument('-v', '--version', action='store_true', help='Outputs the version.')
    parser.add_argument('-q', '--quiet', action='store_true', help='Suppress output from subprocesses.')
    parser.add_argument('-g', '--good', help='The lower bound of the commit range (a good commit)')
    parser.add_argument('-b', '--bad', help='The upper bound of the commit range (a bad commit)')

    args = parser.parse_args()

    SnapshotHoster = SNAPSHOT_HOSTERS[args.hoster]

    if args.version:
        print(pkg_version())
        sys.exit(0)

    opts = Options(
        quiet=args.quiet,
        root_dir=args.root,
        mixxx_dir=args.root / 'mixxx.git',
        mount_dir=args.root / 'mnt',
        log_dir=args.root / 'log',
        downloads_dir=args.root / 'downloads',
    )

    # Ensure that the root directory exists
    opts.root_dir.mkdir(parents=True, exist_ok=True)

    # Clone or fetch repo
    clone_mixxx(opts)

    # Create auxiliary directories
    for dir in [opts.downloads_dir, opts.mount_dir, opts.log_dir]:
        dir.mkdir(parents=True, exist_ok=True)

    # Set up platform-specific snapshot runner
    runner = SnapshotRunner(opts)

    # Fetch snapshots and match them up with Git commits
    hoster = SnapshotHoster(
        branch=args.branch,
        suffix=runner.download_path.suffix,
        opts=opts
    )

    snapshots = hoster.fetch_snapshots()
    if args.dump_snapshots:
        print(json.dumps(snapshots, indent=2))

    commits = sort_commits(list(snapshots.keys()), opts)

    if commits:
        print(f'{len(commits)} snapshot commits found')
    else:
        print('No snapshot commits found (or some error occurred while sorting the commits)')
        sys.exit(1)

    # Parse search range bounds
    good = parse_commit(args.good or commits[0], opts)
    bad = parse_commit(args.bad or commits[-1], opts)

    if good == bad or not commits_in_order([good, bad], opts):
        print('Please make sure that good < bad!')
        sys.exit(1)

    try:
        good_idx = commits.index(good)
    except ValueError:
        print(f'Good commit {good} has no associated snapshot!')
        sys.exit(1)
    try:
        bad_idx = commits.index(bad)
    except ValueError:
        print(f'Bad commit {bad} has no associated snapshot!')
        sys.exit(1)

    # Binary search over the commits
    while good_idx < bad_idx - 1:
        print(f'==> Searching {commits[good_idx][:10]} to {commits[bad_idx][:10]} ({bad_idx - good_idx} commits)...')
        mid_idx = (bad_idx + good_idx) // 2
        mid = commits[mid_idx]

        print(f'==> Checking {describe_commit(mid, opts)}')
        download_snapshot(snapshots[mid], runner.download_path)
        runner.setup_snapshot()
        runner.run_snapshot()
        runner.cleanup_snapshot()

        answer = ''
        while not answer or answer not in 'yn':
            answer = input('Good? [y/n] ')

        if answer == 'y':
            good_idx = mid_idx
        else:
            bad_idx = mid_idx

    print(f'Last good: {describe_commit(commits[good_idx], opts)}')
    print(f'First bad: {describe_commit(commits[bad_idx], opts)}')
    print(f'Trees:     https://github.com/mixxxdj/mixxx/tree/{commits[good_idx]}')
    print(f'           https://github.com/mixxxdj/mixxx/tree/{commits[bad_idx]}')
    print(f'Diff:      https://github.com/mixxxdj/mixxx/compare/{commits[good_idx]}...{commits[bad_idx]}')



from pathlib import Path

from mixxx_bisect.options import Options
from mixxx_bisect.runner import SnapshotRunner
from mixxx_bisect.utils.run import run

import shutil

class WindowsSnapshotRunner(SnapshotRunner):
    def __init__(self, opts: Options):
        self.mount_dir = opts.mount_dir
        self.install_dir = opts.mount_dir / 'Mixxx'
        self.opts = opts
    
    @property
    def download_path(self) -> Path:
        return self.opts.downloads_dir / 'mixxx-current.msi'

    def setup_snapshot(self):
        print('Extracting snapshot...')
        run([
            'msiexec',
            '/a', str(self.download_path),           # Install the msi
            '/q',                                    # Install quietly i.e. without GUI
            f'TARGETDIR={self.opts.mount_dir}',                # Install to custom target dir
            '/li', str(self.opts.log_dir / 'msi-install.log'), # Log installation to file
        ], opts=self.opts)

    def run_snapshot(self):
        print('Running snapshot...')
        run([str(self.install_dir / 'mixxx.exe')], opts=self.opts)

    def cleanup_snapshot(self):
        print('Cleaning up snapshot...')
        for path in self.opts.mount_dir.iterdir():
            if path.is_file():
                path.unlink()
            else:
                shutil.rmtree(path)



from pathlib import Path

from mixxx_bisect.utils.request import download

def download_snapshot(url: str, download_path: Path):
    print(f'Downloading snapshot...')
    download(url, download_path)



from typing import Optional, cast

from mixxx_bisect.hoster import SnapshotHoster
from mixxx_bisect.options import Options
from mixxx_bisect.utils.git import try_parse_commit
from mixxx_bisect.utils.request import get_soup

import re

SNAPSHOTS_BASE_URL = 'https://downloads.mixxx.org/snapshots/'
SNAPSHOT_NAME_PATTERNS = [
    # Old pattern, e.g. mixxx-main-r7715-a0f80e8464
    re.compile(r'^mixxx-[\w\.]+-r\d+-(\w+)$'),
    # New pattern, e.g. mixxx-2.4-alpha-6370-g44f29763ed-macosintel
    re.compile(r'^mixxx-[\d\.]+(?:-[a-z]+)?-\d+-g(\w+)-\w+$'),
]

def parse_commit_from_name(name: str, suffix: str) -> Optional[str]:
    name = name.removesuffix(suffix)
    for pattern in SNAPSHOT_NAME_PATTERNS:
        matches = pattern.search(name)
        if matches:
            return matches[1]
    return None

class MixxxOrgSnapshotHoster(SnapshotHoster):
    def __init__(self, branch: str, suffix: str, opts: Options):
        self.snapshots_url = f'{SNAPSHOTS_BASE_URL}{branch}/'
        self.branch = branch
        self.suffix = suffix
        self.opts = opts
    
    def fetch_snapshots(self) -> dict[str, str]:
        print(f'==> Fetching snapshots from {self.snapshots_url}...')
        snapshot_soup = get_soup(self.snapshots_url)
        links = [cast(str, a.get('href')) for a in snapshot_soup.select('a')]
        commits = [parse_commit_from_name(link.split('/')[-1], self.suffix) for link in links]
        parsed_commits = [try_parse_commit(commit, self.opts) if commit else None for commit in commits]
        return {
            commit: f'{self.snapshots_url}/{link}'
            for commit, link in zip(parsed_commits, links)
            if commit and link.endswith(self.suffix)
        }



from pathlib import Path
from typing import Optional

from mixxx_bisect.options import Options

import subprocess

def run(cmd: list[str], opts: Options, cwd: Optional[Path]=None):
    subprocess.run(
        cmd,
        cwd=cwd or opts.root_dir,
        stdout=subprocess.DEVNULL if opts.quiet else None,
        stderr=subprocess.DEVNULL if opts.quiet else None,
    )

def run_with_output(cmd: list[str], opts: Options, cwd: Optional[Path]=None) -> list[str]:
    result = subprocess.run(
        cmd,
        cwd=cwd or opts.root_dir,
        check=True,
        capture_output=True,
        encoding='utf8',
    )
    return result.stdout.splitlines()



from typing import Optional, cast

from mixxx_bisect.hoster import SnapshotHoster
from mixxx_bisect.options import Options
from mixxx_bisect.utils.git import try_parse_commit
from mixxx_bisect.utils.request import get

import re

RELEASES_API_URL = 'https://api.github.com/repos/fwcd/m1xxx/releases'
SNAPSHOT_NAME_PATTERNS = [
    re.compile(r'^mixxx-[\d\.]+\.c\d+\.r(\w+)$'),
]

def parse_commit_from_name(name: str, suffix: str) -> Optional[str]:
    name = name.removesuffix(suffix)
    for pattern in SNAPSHOT_NAME_PATTERNS:
        matches = pattern.search(name)
        if matches:
            return matches[1]
    return None

class M1xxxSnapshotHoster(SnapshotHoster):
    def __init__(self, branch: str, suffix: str, opts: Options):
        self.suffix = suffix
        self.opts = opts
    
    def fetch_snapshots(self) -> dict[str, str]:
        print(f'==> Fetching snapshots from {RELEASES_API_URL}...')
        # TODO: Proper paging over more than the latest 100 snapshots
        releases = get(f'{RELEASES_API_URL}?per_page=100', json=True).json()
        urls = [
            asset['browser_download_url']
            for release in releases
            for asset in release.get('assets', [])
            if cast(str, asset['name']).endswith(self.suffix)
        ]
        commits = [parse_commit_from_name(url.split('/')[-1], self.suffix) for url in urls]
        parsed_commits = [try_parse_commit(commit, self.opts) if commit else None for commit in commits]
        return {
            commit: url
            for commit, url in zip(parsed_commits, urls)
            if commit and url.endswith(self.suffix)
        }



from typing import Optional

from mixxx_bisect.options import Options
from mixxx_bisect.utils.run import run, run_with_output

import subprocess

def clone_mixxx(opts: Options):
    if opts.mixxx_dir.exists():
        print('==> Fetching Mixxx...')
        run(['git', 'fetch', 'origin'], opts=opts, cwd=opts.mixxx_dir)
    else:
        print('==> Cloning Mixxx...')
        run(['git', 'clone', '--bare', 'https://github.com/mixxxdj/mixxx.git', str(opts.mixxx_dir)], opts=opts)

def sort_commits(commits: list[str], opts: Options) -> list[str]:
    # Windows doesn't like passing too many (long) args, so we truncate commit hashes to 8 chars
    commits = [commit[:8] for commit in commits]
    lines = run_with_output(['git', 'rev-list', '--no-walk'] + commits, cwd=opts.mixxx_dir, opts=opts)
    return lines[::-1]

def parse_commit(rev: str, opts: Options) -> str:
    lines = run_with_output(['git', 'rev-parse', rev], cwd=opts.mixxx_dir, opts=opts)
    return lines[0]

def try_parse_commit(rev: str, opts: Options) -> Optional[str]:
    try:
        return parse_commit(rev, opts)
    except subprocess.CalledProcessError:
        return None

def show_commit(rev: str, format: str, opts: Options) -> str:
    lines = run_with_output(['git', 'show', '-s', f'--format={format}', rev], cwd=opts.mixxx_dir, opts=opts)
    return lines[0]

def describe_commit(rev: str, opts: Options) -> str:
    commit = parse_commit(rev, opts)
    return f"{commit[:10]} from {show_commit(rev, '%ci', opts)} ({show_commit(rev, '%s', opts)})"

def commits_in_order(commits: list[str], opts: Options) -> bool:
    return commits == sort_commits(commits, opts)
